"""
This plugin allows you to use Belsie code in MkDocs documentation.

Use a code block with language `nelsie` in Markdown.
Each slide has access to `slides` and `slide` variables.
You can use steps in the code block.

You can optionally use the following parameters:
- width=<int>: Width of the resulting image.
- height=<int>: Height of the resulting image.
- type=<lib, render>: If `type` is `lib`, do not render the code block as a slide, but use it as
a library function for futher slides on the page.
- debug=<yes, no>: Whether to use debug_boxes.

Example:
```nelsie,width=300,height=300
slide.text("Hello world")
```
"""

import contextlib
import os
import base64
import uuid
from typing import List

from mkdocs.config.config_options import Choice
from mkdocs.plugins import BasePlugin


def is_fence_delimiter(line):
    line = line.strip()
    return line.startswith("```") or line.startswith("~~~")


def get_fence_char(header):
    return "`" if header[0] == "`" else "~"


def parse_fence_header(header):
    fence_char = get_fence_char(header)
    items = header.strip(f" {fence_char}").split(",")
    lang = items[0]

    def get_kv(item):
        vals = item.split("=")
        if len(vals) == 1:
            return (vals[0], "yes")
        return vals

    args = dict(get_kv(item) for item in items[1:])
    return (lang, args)


def nelsie_to_python_header(header):
    fence_char = get_fence_char(header) * 3
    start = header.index(f"{fence_char}nelsie")
    return header[:start] + f"{fence_char}python"


class CodeContext:
    def __init__(self):
        self.lib = []

    def add_lib_code(self, code_lines):
        self.lib += code_lines

    def get_lib_code(self):
        return "\n".join(self.lib)


def trim_indent(lines):
    min_indent = min(len(line) - len(line.lstrip()) for line in lines)
    return [line[min_indent:] for line in lines]


@contextlib.contextmanager
def change_cwd(directory: str):
    cwd = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(cwd)


def render_slide(
    code: List[str],
    docs_dir: str,
    ctx: CodeContext,
    width: int,
    height: int,
    debug_boxes: bool,
    render_format: str,
) -> str:
    code = [line for line in code if "#!IGNORE" not in line]
    code = "\n".join(trim_indent(code))
    template = f"""
from nelsie import SlideDeck, TextStyle, Arrow, Stroke, InSteps, Path
deck = SlideDeck()
{code}
""".strip()

    locals = {}
    code_object = compile(template, "nelsie_render.py", "exec")

    with change_cwd(docs_dir):
        exec(code_object, locals)  # Sorry
    deck = locals["deck"]
    result = deck.render(None, "png")
    uid = uuid.uuid4().hex
    n_pages = len(result)
    button_style = "border: 1px solid black; cursor: pointer; background-color: lightgray; padding: 0.5em; margin: 0.5em; border-radius: 10px;"

    update_js = f"document.getElementById('{uid}_current').textContent = c_{uid}; document.getElementById('{uid}_img').src = data_{uid}[c_{uid} - 1]"
    on_click1 = f"c_{uid} -= 1; if (c_{uid} < 1) c_{uid} = 1; {update_js}"
    on_click2 = f"c_{uid} += 1; if (c_{uid} > {n_pages}) c_{uid} = {n_pages}; {update_js}"
    data_array = [f'"data:image/png;base64,{base64.b64encode(data).decode()}"' for _, _, data in result]
    inital_page = 1

    if n_pages > 1:
        return f"""
        <div>
        <script type='text/javascript'>
        var data_{uid} = [{",".join(data_array)}];
        var c_{uid} = {inital_page};
        </script>
        <div>
        <button type="button" style="{button_style}" onClick="{on_click1}">Prev</button>
        <span id="{uid}_current">1</span>/{n_pages}
        <button type="button" style="{button_style}" onClick="{on_click2}">Next</button>
        </div>
        <img id="{uid}_img" style="border: 1px solid black" width="300"/>
        <script>{update_js}</script>
        </div>
        """.strip()
    else:
        return f"""
        <div style="padding-bottom: 1.5em"><img style="border: 1px solid black" width="300" src={data_array[0]}/></div>
        """.strip()


def iterate_fences(src: str, handle_fence):
    lines = []
    fence_content = []
    inside_fence = False
    fence_header = None
    nested = 0

    for line in src.splitlines(keepends=False):
        if is_fence_delimiter(line):
            if inside_fence:
                # Handle nested fenced code blocks
                if len(line.strip()) != 3:
                    nested += 1
                    fence_content.append(line)
                    continue
                elif nested > 0:
                    nested -= 1
                    fence_content.append(line)
                    continue
                assert nested == 0
                assert fence_header is not None
                header, after, fence_content = handle_fence(fence_header, fence_content)
                lines.append(header)
                lines += fence_content
                lines.append(line)
                lines += after or []

                fence_content = []
                inside_fence = False
                fence_header = None
            else:
                inside_fence = True
                fence_header = line
        elif inside_fence:
            fence_content.append(line)
        else:
            lines.append(line)

    return "\n".join(lines)


def strip_ignore_suffix(lines):
    result = []
    for line in lines:
        if "#!IGNORE" not in line:
            result.append(line)
        else:
            result.append(line[: line.index("#!IGNORE")].rstrip())
    return result


class NelsiePlugin(BasePlugin):
    config_scheme = (("render_format", Choice(("svg", "png"))),)

    def on_page_markdown(self, src: str, page, config, *args, **kwargs):
        # TODO: use Markdown parser

        render_format = self.config.get("render_format", "svg")
        docs_dir = config["docs_dir"]
        ctx = CodeContext()

        def handle_fence(header, fence_lines):
            lang, args = parse_fence_header(header)
            if lang == "nelsie":
                lines = []
                type = args.get("type", "render")
                if type == "lib":
                    ctx.add_lib_code(fence_lines)
                elif type == "render":
                    width = args.get("width", "300")
                    height = args.get("height", "300")
                    debug_boxes = args.get("debug", "no")

                    lines = render_slide(
                        fence_lines,
                        docs_dir,
                        ctx,
                        width=width,
                        height=height,
                        debug_boxes=debug_boxes == "yes",
                        render_format=render_format,
                    ).splitlines(keepends=False)
                    fence_lines = strip_ignore_suffix(fence_lines)
                return nelsie_to_python_header(header), lines, fence_lines
            return header, [], fence_lines

        return iterate_fences(src, handle_fence)



import os
import sys

import pytest

PYTEST_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(PYTEST_DIR)
NELSIE_BIN = os.path.join(ROOT_DIR, "nelsie-builder", "target", "debug", "nelsie-builder")
CHECKS_DIR = os.path.join(PYTEST_DIR, "checks")
ASSETS_DIR = os.path.join(PYTEST_DIR, "assets")

sys.path.insert(0, os.path.join(ROOT_DIR, "nelsie-api"))

from nelsie import Resources, SlideDeck  # noqa


def new_resources(**kwargs):
    resources = Resources(**kwargs)

    # Let us fix the fonts, so we have the same results across all OSs
    resources.load_fonts_dir(os.path.join(ASSETS_DIR, "fonts"))
    return resources


@pytest.fixture(scope="session")
def resources():
    return new_resources()


@pytest.fixture()
def deck_builder(resources):
    def helper(**kwargs):
        kwargs.setdefault("image_directory", ASSETS_DIR)
        kwargs.setdefault("default_font", "DejaVu Sans")
        kwargs.setdefault("default_monospace_font", "DejaVu Sans Mono")
        kwargs.setdefault("resources", resources)
        return SlideDeck(**kwargs)

    return helper


@pytest.fixture()
def deck(deck_builder):
    return deck_builder()



from nelsie import Stroke, Path, TextStyle
from testutils import check

from nelsie.layoutexpr import LayoutExpr


@check()
def test_fix_sizes(deck):
    slide = deck.new_slide()
    slide.box(width="100", height=300, bg_color="red")
    slide.box(width="100%", height="20", bg_color="blue")
    slide.box(width="50%", height="20%", bg_color="green")


@check(n_slides=4)
def test_layout_directions(deck):
    def make_slide(box):
        box.box(width=100, height=100, bg_color="red")
        box.box(width=200, height=200, bg_color="green")
        box.box(width=300, height=300, bg_color="blue")

    make_slide(deck.new_slide().box(width="100%", height="100%", row=False, reverse=False))
    make_slide(deck.new_slide().box(width="100%", height="100%", row=False, reverse=True))
    make_slide(deck.new_slide().box(width="100%", height="100%", row=True, reverse=False))
    make_slide(deck.new_slide().box(width="100%", height="100%", row=True, reverse=True))


@check(n_slides=5)
def test_layout_position(deck):
    def make_test_slide(row=False, **kwargs):
        slide = deck.new_slide(width=400, height=300)
        b = slide.box(row=row, width="50%", height="50%", bg_color="gray")
        b.box(width=20, height=20, bg_color="green")
        b2 = b.box(width=20, height=20, bg_color="red", **kwargs)
        b2.box(width="70%", height="70%", bg_color="orange")
        b.box(width=20, height=20, bg_color="green")

    # Slide 0 - both position defined
    make_test_slide(x=40, y=130)

    # Slide 1 - only x defined
    make_test_slide(x=40)

    # Slide 2 - only y defined
    make_test_slide(y=130)

    # Slide 3 - only x defined (row)
    make_test_slide(x=40, row=True)

    # Slide 4 - only y defined (row)
    make_test_slide(y=130, row=True)


@check()
def test_layout_position_string(deck):
    slide = deck.new_slide(width=400, height=300)
    slide.box(x="30", y="30%", width="50", height="50", bg_color="red")
    slide.box(x=30, y="80%", width="50", height="50", bg_color="blue")


@check()
def test_z_level(deck):
    slide = deck.new_slide(width=100, height=100)
    slide.box(x=0, y=0, width=75, height=75, z_level=1, bg_color="red")
    slide.box(x=25, y=25, width=75, height=75, bg_color="green")


@check()
def test_layout_padding(deck):
    slide = deck.new_slide(width=150, height=150)
    b = slide.box(bg_color="gray", p_left=20)
    b.box(width=15, height=15, bg_color="red")
    b = slide.box(bg_color="gray", p_right=20)
    b.box(width=15, height=15, bg_color="blue")
    b = slide.box(bg_color="gray", p_top=20)
    b.box(width=15, height=15, bg_color="red")
    b = slide.box(bg_color="gray", p_bottom=20)
    b.box(width=15, height=15, bg_color="blue")


@check()
def test_layout_margin(deck):
    slide = deck.new_slide(width=150, height=150)
    b = slide.box(bg_color="gray", m_left=20)
    b.box(width=15, height=15, bg_color="red")
    b = slide.box(bg_color="gray", m_right=20)
    b.box(width=15, height=15, bg_color="blue")
    b = slide.box(bg_color="gray", m_top=20)
    b.box(width=15, height=15, bg_color="red")
    b = slide.box(bg_color="gray", m_bottom=20)
    b.box(width=15, height=15, bg_color="blue")


@check()
def test_layout_flex_grow(deck):
    slide = deck.new_slide(width=150, height=150)
    b = slide.box(height="100%", bg_color="gray")
    b.box(width=30, height=15, bg_color="orange")
    b.box(width=30, flex_grow=1, bg_color="red")
    b.box(width=30, flex_grow=2, bg_color="blue")
    b.box(width=30, height=5, bg_color="#ff00ff")


@check()
def test_layout_flex_wrap(deck):
    slide = deck.new_slide(width=150, height=150)
    b = slide.box(row=True, height="100%", width="100%", bg_color="gray", flex_wrap="wrap")
    b.box(width=40, height=30, bg_color="orange")
    b.box(width=40, height=30, bg_color="blue")
    b.box(width=40, height=30, bg_color="orange")
    b.box(width=40, height=30, bg_color="blue")
    b.box(width=40, height=30, bg_color="orange")
    b.box(width=40, height=30, bg_color="blue")


@check(n_slides=4)
def test_layout_justify_content(deck):
    slide = deck.new_slide(width=50, height=150)
    b = slide.box(height="100%", width="100%", bg_color="gray", justify_content="start")
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")

    slide = deck.new_slide(width=50, height=150)
    b = slide.box(height="100%", width="100%", bg_color="gray", justify_content="end")
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")

    slide = deck.new_slide(width=50, height=150)
    b = slide.box(
        height="100%",
        width="100%",
        bg_color="gray",
        justify_content="space-evenly",
    )
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")

    slide = deck.new_slide(width=50, height=150)
    b = slide.box(
        height="100%",
        width="100%",
        bg_color="gray",
        justify_content="space-around",
    )
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")


@check(n_slides=3)
def test_layout_gap(deck):
    slide = deck.new_slide(width=50, height=150)
    b = slide.box(
        height="100%",
        width="100%",
        bg_color="gray",
        justify_content="start",
        gap=(0, 10),
    )
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")

    slide = deck.new_slide(width=50, height=150)
    b = slide.box(
        height="100%",
        width="100%",
        bg_color="gray",
        justify_content="start",
        gap=(0, "15%"),
    )
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")

    slide = deck.new_slide(width=150, height=50)
    b = slide.box(
        row=True,
        height="100%",
        width="100%",
        bg_color="gray",
        justify_content="start",
        gap=(30, 20),
    )
    b.box(width=20, height=30, bg_color="red")
    b.box(width=20, height=30, bg_color="green")
    b.box(width=20, height=30, bg_color="blue")


def test_layout_expr():
    e = LayoutExpr.x(123) + 10
    assert e._expr == ("sum", ("x", 123), 10)
    e2 = e - 20.0
    assert e2._expr == ("sum", ("x", 123), 10, -20.0)


@check()
def test_expr_x_y_weight_height(deck):
    slide = deck.new_slide(width=200, height=80)
    box = slide.box(width=180, height=40, bg_color="green")
    slide.box(
        x=box.x(0.33),
        y=box.y(),
        width=box.width(0.33),
        height=box.height(0.5),
        bg_color="blue",
    )


@check(n_slides=2)
def test_m_xy_p_xy(deck):
    slide = deck.new_slide(width=100, height=100)
    box = slide.box(bg_color="green")
    box.box(width=30, height=30, m_x=10, m_y=20, bg_color="red")

    slide = deck.new_slide(width=100, height=100)
    box = slide.box(bg_color="green", p_x=20, p_y=10)
    box.box(width=30, height=30, bg_color="red")


@check()
def test_layout_grid_in_pixels(deck):
    slide = deck.new_slide(width=300, height=100)
    box = slide.box(grid_template_rows=[20, 35, 20], grid_template_columns=[40, 100, 100], bg_color="gray")

    box.box(grid_column=1, grid_row=2, bg_color="green")
    box.box(grid_column=2, grid_row=1, bg_color="blue")
    box.box(grid_column=-2, grid_row=(1, "span 3"), bg_color="orange")


@check()
def test_layout_simple_grid_fractions(deck):
    slide = deck.new_slide(width=300, height=100)
    box = slide.box(
        grid_template_rows=["1 fr", "1fr"],
        grid_template_columns=["1 fr", "1 fr"],
        bg_color="gray",
        width="250",
        height="90",
    )
    box.box(grid_column=1, grid_row=1, bg_color="red")


@check()
def test_layout_grid_fractions(deck):
    slide = deck.new_slide(width=300, height=100)
    box = slide.box(
        width="90%",
        height="90%",
        grid_template_rows=["1fr", "2 fr", "1 fr"],
        grid_template_columns=["50%", "1fr", "1fr"],
        bg_color="gray",
    )

    box.box(grid_column=2, grid_row=1, bg_color="green")
    box.box(grid_column=3, grid_row=(1, 3), bg_color="blue")

    box.box(grid_column=1, grid_row=2, bg_color="orange")
    box.box(grid_column=2, grid_row=2, bg_color="black")

    box.box(grid_column=(1, 4), grid_row=3, bg_color="red")


@check()
def test_layout_grid_table(deck):
    data = [
        ("Name", "Time", "Type"),
        ("Jane", 3.5, "A1"),
        ("John", 4.1, "B7"),
        ("Johanna", 12.0, "C1"),
        ("Elise", 12.5, "D4"),
        ("Max", 320.2, "E1"),
    ]

    slide = deck.new_slide(width=1000, height=300)
    table = slide.box(
        width="70%",
        grid_template_columns=["2fr", "1fr", 130],
        grid_template_rows=[50] + [40] * (len(data) - 1),
        bg_color="#ddd",
    )
    header_style = TextStyle(weight=800)
    table.box(grid_column=(1, 4), grid_row=1, bg_color="#fbc")
    for i in range(2, len(data) + 1, 2):
        table.box(grid_column=(1, 4), grid_row=i, bg_color="#eee")
    column1 = table.box(grid_column=2, grid_row=(1, len(data) + 1))
    stroke = Stroke(color="#888", width=2)
    column1.draw(Path(stroke=stroke).move_to(0, 0).line_to(0, "100%"))
    column1.draw(Path(stroke=stroke).move_to("100%", 0).line_to("100%", "100%"))

    # Fill the table with data
    for i, row in enumerate(data, 1):
        s = header_style if i == 1 else None
        table.box(grid_column=1, grid_row=i).text(row[0], s)
        table.box(grid_column=2, grid_row=i, row=True, justify_content="end", m_right=30).text(str(row[1]), s)
        table.box(grid_column=3, grid_row=i, row=True, justify_content="start", m_left=30).text(row[2], s)



from testutils import check


@check()
def test_box_debug_frame(deck):
    slide = deck.new_slide(width=200, height=150)
    slide.box(width=180, height="30%", debug_layout=True, name="Box 123")
    slide.box(width=100, height="30%", debug_layout="orange")
    slide.box(width=100, height="30%")
    slide.box(debug_layout=True)

    # slide = deck.new_slide(width=200, height=150, debug_frames=True)
    # b = slide.box(width="80%", height="80%")
    # b = b.box(width="100%", height="100%")
    # b = b.box(width="99%", height="99%")
    # b.box(width="99%", height="99%")
    #
    # deck.render(output_pdf="/tmp/out.pdf")


@check()
def test_box_border_radius(deck):
    slide = deck.new_slide(width=150, height=150)
    slide.box(x=75, y=20, width=35, height=120, bg_color="red", border_radius=10)
    slide.box(x=10, y=50, width=100, height=10, bg_color="green", border_radius=5)


@check()
def test_pdf_url(deck):
    # Mostly just smoke test, as I do not know how to test annotations in resulting PDF
    slide = deck.new_slide(width=100, height=100)
    slide.box(
        x=65, y=10, width=32, height=20, bg_color="red", border_radius=10, url="https://github.com/spirali/nelsie"
    )



from nelsie import Stroke, Path, Arrow
from testutils import check


@check(n_slides=4)
def test_chessboard(deck):
    slide = deck.new_slide(width=600, height=600)

    # Draw chessboard
    colors = ["orange", "#665566"]
    tiles = {}
    board = slide.box(width=500, height=500)
    for i in range(8):
        row = board.box(width="100%", flex_grow=1.0, row=True)
        for j in range(8):
            b = row.box(height="100%", flex_grow=1.0, bg_color=colors[(i + j) % 2])
            tiles[(j, i)] = b.overlay(z_level=0)

    # Draw arrow
    slide.overlay(show="1-3", z_level=1).draw(
        Path(stroke=Stroke(color="black", width=15), arrow_end=Arrow(30))
        .move_to(tiles[(3, 4)].x(0.5), tiles[(3, 4)].y(0.5))
        .line_to(tiles[(3, 2)].x(0.5), tiles[(3, 2)].y(0.5))
        .line_to(tiles[(4, 2)].x(0.5), tiles[(4, 2)].y(0.5))
    )

    # Draw knight
    tiles[(3, 4)].box(show="1", z_level=2).image("knight.svg")
    tiles[(3, 3)].box(show="2", z_level=2).image("knight.svg")
    tiles[(3, 2)].box(show="3", z_level=2).image("knight.svg")
    tiles[(4, 2)].box(show="4", z_level=2).image("knight.svg")



from typing import Generic, Sequence, TypeVar

T = TypeVar("T")
S = TypeVar("S")

Step = tuple[int]


class InSteps(Generic[T]):
    """
    InSteps is a wrapper that allows to set a different values for each step.
    InSteps defines values in "key" steps, in other steps the value remains until it is changed
    another key step.

    Example:
    ```python
    slide.box(..., bg_color=InSteps({1: "green", 4: "red"})
    ```

    Defines "green" background for steps 1, 2, 3; and "red" for step 4 and further.

    InSteps can be also initialized by a list, then it defines values for first `n` steps,
    where `n` is a length of the list. It means that `InSteps(["a", "b", "c"])` is equal to
    `InSteps({1: "a", 2: "b", 3: "c"})`
    """

    def __init__(
        self,
        values: Sequence[T] | dict[Step | int, T],
    ):
        if isinstance(values, Sequence):
            tmp = {}
            prev = None
            for i, v in enumerate(values):
                if i != 0 and v == prev:
                    continue
                tmp[i + 1] = v
                prev = v
            values = tmp
        elif not isinstance(values, dict):
            raise ValueError("Invalid type for values")
        self.in_step_values = values

    def get(self, step: int, default: S = None) -> T | None:
        v = self.in_step_values.get(step)
        if v is not None:
            return v
        if step <= 0:
            return default
        return self.get(step - 1, default)

    def map(self, fn):
        return InSteps(
            {step: fn(v) for step, v in self.in_step_values.items()},
        )

    def key_steps(self):
        return self.in_step_values.keys()

    def zip(self, other: "InSteps[S]") -> "InSteps[tuple[S, T]]":
        keys = set(self.key_steps())
        keys.update(other.key_steps())
        return InSteps(
            {step: (self.get(step), other.get(step)) for step in keys},
        )


def zip_in_steps(values: list[S | InSteps[S]], default: S) -> InSteps[list[S]]:
    keys = set().union(*[x.key_steps() if isinstance(x, InSteps) else (1,) for x in values])
    return InSteps(
        {step: [x.get(step, default) if isinstance(x, InSteps) else x for x in values] for step in keys},
    )



import os
from dataclasses import dataclass
from typing import Sequence

from .basictypes import (
    AlignItems,
    AlignContent,
    FlexWrap,
    Length,
    LengthAuto,
    Position,
    Size,
    TextAlign,
    parse_debug_layout,
)
from .insteps import InSteps, zip_in_steps
from .layoutexpr import LayoutExpr
from .shapes import Path
from .textstyle import TextStyle, _data_to_text_style
from .watch import watch_path


@dataclass
class TextContent:
    text: str
    style1: TextStyle | str | None
    style2: TextStyle | str | None
    formatting_delimiters: str
    text_align: int
    syntax_language: str | None
    syntax_theme: str | None
    parse_counters: bool


@dataclass
class ImageContent:
    path: str | None | InSteps[str | None]
    enable_steps: bool
    shift_steps: int


NodeContent = ImageContent | TextContent | None
AlignItemsSteps = AlignItems | None | InSteps[AlignItems | None]
AlignContentSteps = AlignContent | None | InSteps[AlignContent | None]
GridTemplate = Sequence[float | str] | InSteps[Sequence[float | str]]
GridPosition = int | str | tuple[int | str] | InSteps[int | str | tuple[int | str]]


class BoxBuilder:
    def get_box(self):
        """
        @private
        """
        raise NotImplementedError

    def set_style(self, name: str, style: TextStyle | InSteps[TextStyle]):
        """
        Set text style under given name.
        """
        box = self.get_box()
        deck = box.deck
        deck._deck.set_style(deck.resources, name, style, False, box.slide._slide_id, box._box_id)

    def update_style(self, name: str, style: TextStyle | InSteps[TextStyle]):
        """
        Load a text style and merge it with the `style` and save it back.
        Throws an exception if a style under given name does not exist.
        """
        box = self.get_box()
        deck = box.deck
        deck._deck.set_style(deck.resources, name, style, True, box.slide._slide_id, box._box_id)

    def get_style(self, name: str, step: int = 1) -> TextStyle:
        """
        Get text style under given name. Style is returned for a given `step`.
        Throws an exception if a style under given name does not exist.

        This function returns text style even when InSteps was used to set the style,
        as it returns text style for a given step.
        """

        box = self.get_box()
        return _data_to_text_style(box.deck._deck.get_style(name, step, box.slide._slide_id, box._box_id))

    def image(self, path: str | None | InSteps[str | None], enable_steps=True, shift_steps=0, **box_args):
        """
        Create a box with an image. Supported formats: SVG, PNG, JPEG, GIF, ORA
        """
        assert shift_steps >= 0

        def process_path(p):
            if p is None:
                return None
            if slide.image_directory is not None:
                p = os.path.join(slide.image_directory, p)
            p = os.path.abspath(p)
            watch_path(p)
            return p

        slide = self.get_box().slide
        if isinstance(path, InSteps):
            path = path.map(process_path)
        else:
            path = process_path(path)
        image = ImageContent(
            path=path,
            enable_steps=enable_steps,
            shift_steps=shift_steps,
        )
        return self.box(_content=image, **box_args)

    def text(
        self,
        text: str | InSteps[str],
        style: str | TextStyle | InSteps[TextStyle] | None = None,
        *,
        parse_styles: bool = True,
        style_delimiters: str | None = "~{}",
        tab_width: int = 4,
        align: TextAlign = "start",
        strip=True,
        parse_counters: bool = False,
        **box_args,
    ):
        """
        Create a box with a text.
        """
        box_args.setdefault("name", "text")
        return self._text_box(
            text,
            style,
            None,
            style_delimiters if parse_styles else None,
            tab_width,
            box_args,
            align,
            None,
            None,
            strip,
            parse_counters,
        )

    def code(
        self,
        text: str | InSteps[str],
        language: str | None = "default",
        style: str | TextStyle | InSteps[TextStyle] | None = None,
        *,
        theme: str | None = None,
        parse_styles: bool = False,
        style_delimiters: str | None = "~{}",
        tab_width: int = 4,
        align: TextAlign = "start",
        strip: bool = True,
        parse_counters: bool = False,
        **box_args,
    ):
        """
        Create a box with a syntax highlighted text.
        """
        box_args.setdefault("name", "code")
        if theme is None:
            theme = self.get_box().deck.default_code_theme
        if language == "default":
            language = self.get_box().deck.default_code_language
        return self._text_box(
            text,
            "code",
            style,
            style_delimiters if parse_styles else None,
            tab_width,
            box_args,
            align,
            language,
            theme,
            strip,
            parse_counters,
        )

    def _text_box(
        self,
        text,
        style1,
        style2,
        delimiters,
        tab_width,
        box_args,
        align,
        language,
        theme,
        strip,
        parse_counters,
    ):
        def text_preprocess(x):
            if strip:
                x = x.strip()
            return x.replace("\t", " " * tab_width)

        if isinstance(text, str):
            text = text_preprocess(text)
        elif isinstance(text, list):
            text = zip_in_steps(text, "").map(lambda x: text_preprocess("".join(x)))
        elif isinstance(text, InSteps):
            text = text.map(text_preprocess)
        else:
            raise Exception("Invalid type for text")

        if align == "start":
            align = 0
        elif align == "center":
            align = 1
        elif align == "end":
            align = 2
        else:
            raise Exception(f"Invalid alignment: {align}")
        text_content = TextContent(
            text=text,
            style1=style1,
            style2=style2,
            formatting_delimiters=delimiters,
            text_align=align,
            syntax_language=language,
            syntax_theme=theme,
            parse_counters=parse_counters,
        )
        return self.box(_content=text_content, **box_args)

    def draw(self, paths: Path | list[Path] | InSteps[Path | list[Path]]):
        """
        Draw one or paths in the slide.
        """

        if isinstance(paths, Path):
            paths = [paths]
        elif isinstance(paths, InSteps):
            paths = paths.map(lambda p: [p] if isinstance(p, Path) else p)
        box = self.get_box()
        box.deck._deck.draw(box.slide._slide_id, box._box_id, paths)

    def box(
        self,
        *,
        active: bool | str | int | InSteps[bool] = True,
        show: bool | str | int | InSteps[bool] = True,
        z_level: int | InSteps[int] | None = None,
        x: Position | InSteps[Position] = None,
        y: Position | InSteps[Position] = None,
        width: Size | InSteps[Size] = None,
        height: Size | InSteps[Size] = None,
        border_radius: float | InSteps[float] = 0,
        p_left: Length | InSteps[Length] | None = None,
        p_right: Length | InSteps[Length] | None = None,
        p_top: Length | InSteps[Length] | None = None,
        p_bottom: Length | InSteps[Length] | None = None,
        p_x: Length | InSteps[Length] = 0,
        p_y: Length | InSteps[Length] = 0,
        m_left: LengthAuto | InSteps[LengthAuto] | None = None,
        m_right: LengthAuto | InSteps[LengthAuto] | None = None,
        m_top: LengthAuto | InSteps[LengthAuto] | None = None,
        m_bottom: LengthAuto | InSteps[LengthAuto] | None = None,
        m_x: LengthAuto | InSteps[LengthAuto] = 0,
        m_y: LengthAuto | InSteps[LengthAuto] = 0,
        row: bool | InSteps[bool] = False,
        reverse: bool | InSteps[bool] = False,
        flex_wrap: FlexWrap | InSteps[FlexWrap] = "nowrap",
        flex_grow: float | InSteps[float] = 0.0,
        flex_shrink: float | InSteps[float] = 1.0,
        align_items: AlignItemsSteps = None,
        align_self: AlignItemsSteps = None,
        justify_self: AlignItemsSteps = None,
        align_content: AlignContentSteps = None,
        justify_content: AlignContentSteps = None,
        gap: tuple[Length, Length] | InSteps[tuple[Length, Length]] = (0.0, 0.0),
        grid_template_rows: GridTemplate = (),
        grid_template_columns: GridTemplate = (),
        grid_row: GridPosition = "auto",
        grid_column: GridPosition = "auto",
        bg_color: str | None | InSteps[str | None] = None,
        url: None | str | InSteps[None | str] = None,
        name: str = "",
        debug_layout: bool | str | None = None,
        replace_steps: dict[int, int] | None = None,
        _content: NodeContent | InSteps[NodeContent] = None,
    ):
        """
        Create a new child box. See [Box reference](https://spirali.github.io/nelsie/guide/box/) for documentation
        """
        parent_box = self.get_box()
        if debug_layout is None:
            debug_layout = parent_box.slide.debug_layout
        else:
            debug_layout = parse_debug_layout(debug_layout)
        if z_level is None:
            z_level = parent_box.z_level

        if p_left is None:
            p_left = p_x
        if p_right is None:
            p_right = p_x
        if p_bottom is None:
            p_bottom = p_y
        if p_top is None:
            p_top = p_y

        if m_left is None:
            m_left = m_x
        if m_right is None:
            m_right = m_x
        if m_bottom is None:
            m_bottom = m_y
        if m_top is None:
            m_top = m_y

        deck = parent_box.deck
        box_id, node_id = deck._deck.new_box(
            deck.resources,
            parent_box.slide._slide_id,
            parent_box._box_id,
            active=active,
            show=show,
            z_level=z_level,
            x=x,
            y=y,
            width=width,
            height=height,
            border_radius=border_radius,
            p_left=p_left,
            p_right=p_right,
            p_top=p_top,
            p_bottom=p_bottom,
            m_left=m_left,
            m_right=m_right,
            m_top=m_top,
            m_bottom=m_bottom,
            row=row,
            reverse=reverse,
            flex_wrap=flex_wrap,
            flex_grow=flex_grow,
            flex_shrink=flex_shrink,
            align_items=align_items,
            align_self=align_self,
            justify_self=justify_self,
            align_content=align_content,
            justify_content=justify_content,
            gap=gap,
            grid_template_rows=grid_template_rows,
            grid_template_columns=grid_template_columns,
            grid_row=grid_row,
            grid_column=grid_column,
            bg_color=bg_color,
            url=url,
            name=name,
            debug_layout=debug_layout,
            replace_steps=replace_steps,
            content=_content,
        )
        return Box(deck, parent_box.slide, box_id, node_id, name, z_level)

    def overlay(self, **box_args):
        """
        Create a new box that spans over the box
        """
        box_args.setdefault("x", 0)
        box_args.setdefault("y", 0)
        box_args.setdefault("width", "100%")
        box_args.setdefault("height", "100%")
        return self.box(**box_args)

    def line_box(self, line_idx: int, **box_args):
        """
        Creates a new box over a text line in the box
        """
        return self.box(
            x=self.line_x(line_idx),
            y=self.line_y(line_idx),
            width=self.line_width(line_idx),
            height=self.line_height(line_idx),
            **box_args,
        )

    def text_anchor_box(self, anchor_id: int, **box_args):
        """
        Creates a new box over a text anchor in the box
        """

        return self.box(
            x=self.text_anchor_x(anchor_id),
            y=self.text_anchor_y(anchor_id),
            width=self.text_anchor_width(anchor_id),
            height=self.text_anchor_height(anchor_id),
            **box_args,
        )

    def x(self, width_fraction: float | int | None = None) -> LayoutExpr:
        """
        Get an expression with X coordinate relative to the box.
        """

        node_id = self.get_box().node_id
        expr = LayoutExpr.x(node_id)
        if width_fraction is None:
            return expr
        return expr + LayoutExpr.width(node_id, width_fraction)

    def y(self, height_fraction: float | int | None = None) -> LayoutExpr:
        """
        Get an expression with Y coordinate relative to the box.
        """
        node_id = self.get_box().node_id
        expr = LayoutExpr.y(node_id)
        if height_fraction is None:
            return expr
        return expr + LayoutExpr.height(node_id, height_fraction)

    def width(self, fraction: float | int = 1.0) -> LayoutExpr:
        """
        Get an expression with width of the parent box.
        """
        node_id = self.get_box().node_id
        return LayoutExpr.width(node_id, fraction)

    def height(self, fraction: float | int = 1.0) -> LayoutExpr:
        """
        Get an expression with height of the parent box.
        """
        node_id = self.get_box().node_id
        return LayoutExpr.height(node_id, fraction)

    def line_x(self, line_idx: int, width_fraction: float | int | None = None) -> LayoutExpr:
        """
        Get an expression with X coordinate of a given line of text in the box.
        """
        node_id = self.get_box().node_id
        expr = LayoutExpr.line_x(node_id, line_idx)
        if width_fraction is None:
            return expr
        return expr + LayoutExpr.line_width(node_id, line_idx, width_fraction)

    def line_y(self, line_idx: int, height_fraction: float | int | None = None) -> LayoutExpr:
        """
        Get an expression with Y coordinate of a given line of text in the box.
        """
        node_id = self.get_box().node_id
        expr = LayoutExpr.line_y(node_id, line_idx)
        if height_fraction is None:
            return expr
        return expr + LayoutExpr.line_height(node_id, line_idx, height_fraction)

    def line_width(self, line_idx: int, fraction: float | int = 1.0) -> LayoutExpr:
        """
        Get an expression with a width of a given line of text in the box.
        """
        node_id = self.get_box().node_id
        return LayoutExpr.line_width(node_id, line_idx, fraction)

    def line_height(self, line_idx: int, fraction: float | int = 1.0) -> LayoutExpr:
        """
        Get an expression with a height of a given line of text in the box.
        """
        node_id = self.get_box().node_id
        return LayoutExpr.line_height(node_id, line_idx, fraction)

    def text_anchor_x(self, anchor_id: int, width_fraction: float | int | None = None) -> LayoutExpr:
        """
        Get an expression with X coordinate of a given text anchor in the box.
        """
        node_id = self.get_box().node_id
        expr = LayoutExpr.text_anchor_x(node_id, anchor_id)
        if width_fraction is None:
            return expr
        return expr + LayoutExpr.text_anchor_width(node_id, anchor_id, width_fraction)

    def text_anchor_y(self, anchor_id: int, height_fraction: float | int | None = None) -> LayoutExpr:
        """
        Get an expression with Y coordinate of a given text anchor in the box.
        """
        node_id = self.get_box().node_id
        expr = LayoutExpr.text_anchor_y(node_id, anchor_id)
        if height_fraction is None:
            return expr
        return expr + LayoutExpr.text_anchor_height(node_id, anchor_id, height_fraction)

    def text_anchor_width(self, anchor_id: int, fraction: float | int = 1.0) -> LayoutExpr:
        """
        Get an expression with a height of a given text anchor in the box.
        """
        node_id = self.get_box().node_id
        return LayoutExpr.text_anchor_width(node_id, anchor_id, fraction)

    def text_anchor_height(self, anchor_id: int, fraction: float | int = 1.0) -> LayoutExpr:
        """
        Get an expression with a height of a given text anchor in the box.
        """
        node_id = self.get_box().node_id
        return LayoutExpr.text_anchor_height(node_id, anchor_id, fraction)


class Box(BoxBuilder):
    """
    The box in slide layout.

    It should be created via calling method `.box()` on a slide or another box.
    Note that interesting methods came from Box's parent: BoxBuilder.
    """

    def __init__(
        self,
        deck,
        slide,
        box_id,
        node_id,
        name: str,
        z_level: int,
    ):
        """
        @private
        """
        self.deck = deck
        self.slide = slide
        self._box_id = box_id
        self.node_id = node_id
        self.name = name
        self.z_level = z_level

    def get_box(self):
        """
        @private
        """
        return self



from testutils import check


def test_render_outputs(tmp_path, deck):
    out_svg = tmp_path / "svgs"
    out_png = tmp_path / "pngs"
    out_pdf = tmp_path / "out.pdf"

    slide = deck.new_slide()
    slide.box(width=400, height=200, bg_color="blue")
    slide.box(width=200, height=400, bg_color="red")

    slide = deck.new_slide(bg_color="green")
    slide.box(width=None, height=None, bg_color="black")

    svg_data = deck.render(None, "svg")
    assert len(svg_data) == 2

    assert deck.render(out_svg, "svg") is None

    with open(out_svg / "0.svg", "rb") as f:
        data = f.read()
        assert data.startswith(b"<svg ")
        assert svg_data[0] == (0, (1,), data)

    with open(out_svg / "1.svg", "rb") as f:
        data = f.read()
        assert data.startswith(b"<svg ")
        assert svg_data[1] == (1, (1,), data)

    png_data = deck.render(None, "png")
    assert len(png_data) == 2

    deck.render(out_png, "png")

    with open(out_png / "0.png", "rb") as f:
        data = f.read()
        assert data[:4] == b"\x89PNG"
        assert png_data[0] == (0, (1,), data)

    with open(out_png / "1.png", "rb") as f:
        data = f.read()
        assert data[:4] == b"\x89PNG"
        assert png_data[1] == (1, (1,), data)

    deck.render(out_pdf, "pdf")

    with open(out_pdf, "rb") as f:
        data = f.read(4)
        assert data == b"%PDF"


@check(n_slides=2)
def test_slide_decorator(deck):
    @deck.slide()
    def my_slide(slide):
        slide.box(width=400, height=200, bg_color="blue")
        slide.box(width=200, height=400, bg_color="red")

    @deck.slide(width=300, height=120, debug_layout="green")
    def my_slide2(slide):
        slide.text("Hello world!")



from dataclasses import dataclass, InitVar
from enum import IntEnum
from typing import Literal

from nelsie.basictypes import Stroke
from nelsie.utils import unpack_dataclass


class FontStretch(IntEnum):
    UltraCondensed = 1
    ExtraCondensed = 2
    Condensed = 3
    SemiCondensed = 4
    Normal = 5
    SemiExpanded = 6
    Expanded = 7
    ExtraExpanded = 8
    UltraExpanded = 9


@dataclass(frozen=True)
class TextStyle:
    font_family: str | list[str] | None = None
    color: str | Literal["empty"] | None = None
    stroke: Stroke | Literal["empty"] | None = None
    size: float | None = None
    line_spacing: float | None = None
    italic: bool | None = None
    stretch: FontStretch | None = None
    underline: bool | None = None
    overline: bool | None = None
    line_through: bool | None = None

    # 1-1000; 400 = Normal, 700 = Bold
    weight: int | None = None

    # Init only fields
    # These are used as helpers for initializing
    # commonly used attributes
    bold: InitVar[bool | None] = None

    def __post_init__(self, bold: bool | None):
        if self.size is not None:
            assert self.size >= 0
        if self.line_spacing is not None:
            assert self.line_spacing >= 0
        if self.weight is not None:
            assert 1 <= self.weight <= 1000
            if bold is not None:
                raise Exception("Cannot set both `weight` and `bold` when creating a TextStyle")
        if bold is not None:
            # Workaround to set frozen attribute
            super().__setattr__("weight", 700)

    def merge(self, other: "TextStyle") -> "TextStyle":
        assert isinstance(other, TextStyle)
        return TextStyle(
            *[b if b is not None else a for (a, b) in zip(unpack_dataclass(self), unpack_dataclass(other))]
        )


def _data_to_text_style(data):
    stretch = data.get("stretch")
    if stretch is not None:
        data["stretch"] = FontStretch(stretch)
    stroke = data.get("stroke")
    if isinstance(stroke, dict):
        data["stroke"] = Stroke(**stroke)
    return TextStyle(**data)



import contextlib
import os
import sys

import pytest
import fitz
from conftest import CHECKS_DIR
from PIL import Image, ImageChops, ImageStat

DEFAULT_THRESHOLD = 105.0


@contextlib.contextmanager
def change_workdir(workdir):
    cwd = os.getcwd()
    try:
        os.chdir(workdir)
        yield
    finally:
        os.chdir(cwd)


def concat_images(imgs):
    dst = Image.new("RGBA", (sum([im.width for im in imgs]), imgs[0].height))
    x = 0
    for im in imgs:
        dst.paste(im, (x, 0))
        x += im.width
    return dst


def compare_images(new_dir, old_dir, n_slides, threshold, resize=False):
    new_names = os.listdir(new_dir)
    new_names.sort()
    if len(new_names) != n_slides:
        raise Exception(f"Expected to produce {n_slides} but {len(new_names)} produced")
    try:
        old_names = os.listdir(old_dir)
    except FileNotFoundError:
        raise Exception(f"Checks {old_dir} does not exists")
    old_names.sort()
    if new_names != old_names:
        raise Exception(f"Produced files do not match with check files; new = {new_names}, old = {old_names}")
    for name1, name2 in zip(new_names, old_names):
        new_img = Image.open(os.path.join(new_dir, name1))
        old_img = Image.open(os.path.join(old_dir, name2))
        if resize:
            new_img = new_img.resize(old_img.size)
            new_img = new_img.convert(old_img.mode)
        difference = ImageChops.difference(new_img, old_img)
        stat = ImageStat.Stat(difference)
        diff = sum(stat.sum) / 255.0
        if diff > threshold:
            combined = concat_images([new_img, old_img, difference])
            path = os.path.abspath(f"combined-{name1}.png")
            combined.save(path)
            raise Exception(f"Slide {os.path.join(new_dir, name1)} difference is {diff} (limit is {threshold})")


def check(
    n_slides: int = 1,
    error=None,
    error_match: str | None = None,
    windows_threshold=None,
    deck_kwargs=None,
):
    def wrapper(fn):
        name = fn.__name__
        if name.startswith("test_"):
            name = name[5:]

        def helper(tmp_path, deck_builder):
            with change_workdir(tmp_path):
                if deck_kwargs is None:
                    deck = deck_builder()
                else:
                    deck = deck_builder(**deck_kwargs)
                fn(deck)
                if error is not None:
                    with pytest.raises(error, match=error_match):
                        deck.render("png", "png")
                else:
                    deck.render("png", "png")
                    with open(os.path.join(tmp_path, "check.txt"), "w") as f:
                        f.write(name)
                    deck.render(os.path.join(tmp_path, "output.pdf"), "pdf")
                    os.mkdir("pdf2png")

                    doc = fitz.open("output.pdf")
                    for i, page in enumerate(doc.pages()):
                        pixmap = page.get_pixmap()
                        img = pixmap.tobytes()
                        with open(os.path.join("pdf2png", f"{i}.png"), "wb") as f:
                            f.write(img)
                    threshold = DEFAULT_THRESHOLD
                    if windows_threshold is not None and sys.platform == "win32":
                        threshold = windows_threshold
                    compare_images(
                        os.path.join(tmp_path, "png"),
                        os.path.join(CHECKS_DIR, "png", name),
                        n_slides,
                        threshold=threshold,
                    )
                    compare_images(
                        os.path.join(tmp_path, "pdf2png"),
                        os.path.join(CHECKS_DIR, "pdf2png", name),
                        n_slides,
                        threshold=threshold,
                    )

        return helper

    return wrapper



import typing
from dataclasses import dataclass
from typing import Literal

from .layoutexpr import LayoutExpr

Length = int | float | str
LengthAuto = Length | Literal["auto"]
Size = Length | None
Position = Length | None | LayoutExpr


@dataclass
class Stroke:
    color: str
    width: float = 1.0
    dash_array: list[float] | None = None
    dash_offset: float = 0.0


TextAlign = Literal["start", "center", "end"]


AlignItems = Literal[
    "start",
    "end",
    "flex-start",
    "flex-end",
    "flex-end",
    "center",
    "stretch",
    "baseline",
]

AlignContent = Literal[
    "start",
    "end",
    "flex-start",
    "flex-end",
    "flex-end",
    "center",
    "stretch",
    "space-between",
    "space-evenly",
    "space-around",
]

FlexWrap = Literal["nowrap", "wrap", "wrap-reverse"]


def parse_debug_layout(value: bool | str) -> str | None:
    if value is True:
        return "#ff00ff"
    elif not value:
        return None
    return typing.cast(str, value)



import base64
import getpass
import shutil
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
import os
import click
from PIL import Image, ImageChops, ImageStat
import numpy as np

ROOT = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
CHECKS = os.path.join(ROOT, "tests", "checks")

LIMIT = 105.0


def image_diff(path1, path2, resize):
    try:
        new_img = Image.open(path1)
        old_img = Image.open(path2)
    except FileNotFoundError:
        return float("inf"), None

    if resize:
        new_img = new_img.resize(old_img.size)
        new_img = new_img.convert(old_img.mode)
    difference = ImageChops.difference(new_img, old_img)
    stat = ImageStat.Stat(difference)
    s = sum(stat.sum) / 255.0
    if s > 0:
        sub = np.subtract(np.array(new_img).astype(np.cfloat), np.array(old_img).astype(np.cfloat))
        diff_data2 = -np.minimum(np.min(sub, axis=2), 0)
        diff_data1 = np.maximum(np.max(sub, axis=2), 0)
        diff_data3 = np.zeros_like(diff_data2)
        diff_data = np.stack([diff_data1, diff_data2, diff_data3], axis=2)
        diff_img = Image.fromarray(diff_data.astype(np.uint8), "RGB")
    else:
        diff_img = None
    return s, diff_img


@dataclass
class FailedTest:
    name: str
    files: list[str]
    diffs: list[float]
    test_dir: str
    check_dir: str
    diff_imgs: list
    test_type: str


def check_images(name, path, fails, subdir, resize=False):
    test_dir = os.path.join(os.path.dirname(path), subdir)
    check_dir = os.path.join(CHECKS, subdir, name)
    files = set()
    files.update(os.listdir(test_dir))
    if os.path.exists(check_dir):
        files.update(os.listdir(check_dir))
    files = sorted(files)
    diffs = []
    diff_imgs = []
    for file in files:
        diff, diff_img = image_diff(os.path.join(test_dir, file), os.path.join(check_dir, file), resize)
        diffs.append(diff)
        diff_imgs.append(diff_img)
    if max(diffs) > LIMIT:
        fails.append(
            FailedTest(
                name=name,
                files=files,
                diffs=diffs,
                test_dir=str(test_dir),
                check_dir=check_dir,
                diff_imgs=diff_imgs,
                test_type=subdir,
            )
        )


def collect_failed_tests(path, version) -> list[FailedTest]:
    if path is None:
        path = f"/tmp/pytest-of-{getpass.getuser()}/pytest-current/"
    paths = Path(path).rglob(f"*{version}/check.txt")
    fails = []
    count = 0
    for path in paths:
        count += 1
        with open(path, "r") as f:
            name = f.read()
        check_images(name, path, fails, "png")
        check_images(name, path, fails, "pdf2png", resize=True)
    return fails, count


def html_inline_img(img):
    diff_img_png = BytesIO()
    img.save(diff_img_png, format="PNG")
    img_str = base64.b64encode(diff_img_png.getvalue())
    return f"data:image/png;base64,{img_str.decode()}"


def report(failed_tests):
    filename = "report.html"
    img_dir = os.path.abspath("report_imgs")
    style = "border: 3px solid black"
    shutil.rmtree(img_dir, ignore_errors=True)
    os.mkdir(img_dir)
    for failed_test in failed_tests:
        shutil.copytree(
            os.path.join(failed_test.test_dir),
            os.path.join("report_imgs", failed_test.test_type, failed_test.name),
        )
    with open(filename, "w") as f:
        f.write("<html><body><h1>Test report</h1>")
        for failed_test in failed_tests:
            f.write(f"<h2>{failed_test.name} ({failed_test.test_type})</h2>")
            f.write("<table>")
            f.write("<tr><td>Name</td><td>Reference</td><td>Test</td><td>Diff</td></tr>")
            for file, df, diff_img in zip(failed_test.files, failed_test.diffs, failed_test.diff_imgs):
                f.write(f"<tr><td>{file}</td>")
                f.write(
                    f"<td><img style='{style}' width='400' src='{os.path.join(failed_test.check_dir, file)}'/></td>"
                )
                f.write(
                    f"<td><img style='{style}' width='400' src='{os.path.join(img_dir, failed_test.test_type, failed_test.name, file)}'/></td>"
                )
                if diff_img is not None:
                    f.write(f"<td><img style='{style}' width='400' src='{html_inline_img(diff_img)}'/></td>")
                else:
                    f.write("<td></td>")
                f.write(f"<td>{df:0.2f}</td>")
                f.write("</tr>")
            f.write("</table>")
        f.write("</body></html>")
    print(f"Report written into '{filename}'")


def update(failed_tests):
    for failed_test in failed_tests:
        print(f"Updating {failed_test.name} ({failed_test.test_type})")
        shutil.rmtree(failed_test.check_dir, ignore_errors=True)
        shutil.copytree(
            failed_test.test_dir,
            failed_test.check_dir,
        )


@click.command()
@click.option("--do-update/--do-not-update", default=False)
@click.option("--path", default=None)
@click.option("--version", default="current")
def main(do_update, path, version):
    failed_tests, count = collect_failed_tests(path, version)
    if count == 0:
        print("No test found")
    elif failed_tests:
        if not do_update:
            print("Failed tests:")
            for failed_test in failed_tests:
                print(failed_test.test_type, failed_test.name)
            report(failed_tests)
        else:
            update(failed_tests)
    else:
        print(f"All tests ({count}) are ok")


if __name__ == "__main__":
    main()



import pathlib
from typing import Literal, List

from . import nelsie as nelsie_rs
from .basictypes import parse_debug_layout
from .box import Box, BoxBuilder
from .insteps import Step
from .textstyle import TextStyle, _data_to_text_style

Resources = nelsie_rs.Resources


class Slide(BoxBuilder):
    """
    Represents a slide in a slide deck.

    It should be created via calling method `.new_slide()` on a deck via decorator `@deck.slide()`
    """

    def __init__(
        self,
        deck: "SlideDeck",
        slide_id: int,
        name: str,
        image_directory: str,
        debug_layout: str | None,
    ):
        """
        @private
        """
        self.deck = deck
        self._slide_id = slide_id
        self.name = name
        self.image_directory = image_directory
        self.debug_layout = debug_layout
        self.root_box = Box(deck, self, [], 0, name, 0)

    def get_box(self):
        """
        @private
        """
        return self.root_box

    def insert_step(self, value: Step):
        self.deck._deck.insert_step(self._slide_id, value)

    def remove_step(self, value: Step):
        self.deck._deck.remove_step(self._slide_id, value)

    def remove_steps_below(self, value: Step):
        self.deck._deck.remove_steps_below(self._slide_id, value)

    def remove_steps_above(self, value: Step):
        self.deck._deck.remove_steps_above(self._slide_id, value)

    def set_steps(self, values: set[Step] | frozenset[Step]):
        self.deck._deck.set_steps(self._slide_id, values)

    def get_steps(self) -> list[Step]:
        return self.deck._deck.get_steps(self._slide_id)

    def new_slide_at(self, step: int, **slide_kwargs):
        slide_kwargs["parent_slide"] = (self._slide_id, step)
        return self.deck.new_slide(**slide_kwargs)

    def slide_at(self, step: int, **slide_kwargs):
        slide_kwargs["parent_slide"] = (self._slide_id, step)
        return self.deck.slide(**slide_kwargs)


class SlideDeck:
    def __init__(
        self,
        *,
        width: float = 1024,
        height: float = 768,
        bg_color: str = "white",
        image_directory: str | None = None,
        resources: Resources | None = None,
        default_font: str | None = None,
        default_monospace_font: str | None = None,
        default_code_theme: str = "InspiredGitHub",
        default_code_language: str | None = None,
    ):
        """
        A top-level class of Nelsie. It represents a set of slides.

        Arguments:
        * width - default width of a slide (could be overriden for each slide)
        * height - default width of a slide (could be overriden for each slide)
        * bg_color - default background color a slide (could be overriden for each slide)
        * image_directory - default path where images are searched for (could be overriden for each slide)
        * resource - Resource instance, if None a new instance is created
        * default_font - Name of default font
        * default_monospace_font - Name of the default monospace font
        * default_code_theme - Name of default theme for syntax highlighting (.code() method):
            Available themes:
            * "base16-ocean.dark"
            * "base16-eighties.dark"
            * "base16-mocha.dark"
            * "base16-ocean.light"
            * "InspiredGitHub"
            * "Solarized (dark)"
            * "Solarized (light)"
        * default_code_language - Default language to use for syntax highlighting (.code() method)
        """
        if resources is None:
            resources = Resources()

        self.width = width
        self.height = height
        self.bg_color = bg_color
        self.image_directory = image_directory
        self.resources = resources
        self.default_code_theme = default_code_theme
        self.default_code_language = default_code_language
        self._deck = nelsie_rs.Deck(resources, default_font, default_monospace_font)
        self._slides: List[Slide] = []

    def set_style(self, name: str, style: TextStyle):
        self._deck.set_style(self.resources, name, style, False, None, None)

    def update_style(self, name: str, style: TextStyle):
        self._deck.set_style(self.resources, name, style, True, None, None)

    def get_style(self, name: str, step: int = 1) -> TextStyle:
        return _data_to_text_style(self._deck.get_style(name, step, None, None))

    def new_slide(
        self,
        *,
        width: float | None = None,
        height: float | None = None,
        bg_color: str | None = None,
        image_directory: str | None = None,
        name: str = "",
        debug_steps: bool = False,
        debug_layout: bool | str = False,
        counters: list[str] | None = None,
        parent_slide: tuple[Slide, int] | None = None,
        step_1: bool = True,
    ) -> Slide:
        """
        Creates a new slide in the slide deck.
        """
        if width is None:
            width = self.width
        if height is None:
            height = self.height
        if bg_color is None:
            bg_color = self.bg_color
        if image_directory is None:
            image_directory = self.image_directory
        debug_layout = parse_debug_layout(debug_layout)
        slide_id = self._deck.new_slide(width, height, bg_color, name, step_1, debug_steps, counters, parent_slide)

        slide = Slide(self, slide_id, name, image_directory, debug_layout)
        self._slides.append(slide)
        return slide

    def slide(
        self,
        *,
        width: float | None = None,
        height: float | None = None,
        bg_color: str | None = None,
        image_directory: str | None = None,
        name: str = "",
        debug_steps: bool = False,
        debug_layout: bool | str = False,
        counters: list[str] | None = None,
        parent_slide: tuple[Slide, int] | None = None,
        step_1: bool = True,
    ):
        """
        Decorator for creating new slide.
        It immediately calls the decorated function that should define content of the slide.
        Slide is automatically added into the deck.

        Example:
        ```python
        deck = SlideDeck()

        @deck.slide()
        def my_first_slide(slide):
            slide.text("Hello!")
        ```
        """

        def helper(fn):
            slide = self.new_slide(
                width=width,
                height=height,
                bg_color=bg_color,
                image_directory=image_directory,
                name=name,
                debug_steps=debug_steps,
                debug_layout=debug_layout,
                counters=counters,
                parent_slide=parent_slide,
                step_1=step_1,
            )
            fn(slide)
            return slide

        return helper

    def render(
        self,
        path: str | pathlib.Path | None,
        output_format: Literal["pdf"] | Literal["svg"] | Literal["png"] = "pdf",
        *,
        verbose: int = 1,
        n_threads: int | None = None,
    ) -> None | list[bytes]:
        """
        Render slides

        If format is "pdf" then a single PDF file is created. If format is "svg" or "png" then
        `path` specifies a directory where the slides are created as an individual files.

        If `path` is None then objects are not written to the file system, and they are returned as python objects
        from the method call.
        """
        if path:
            path = str(path)
        return self._deck.render(self.resources, verbose, output_format, path, n_threads)



from conftest import new_resources, ASSETS_DIR

from nelsie import SlideDeck


def test_resources_syntaxes(resources):
    syntaxes = resources.syntaxes()
    s = dict(syntaxes)
    assert "Python" in s
    assert "Rust" in s
    assert "C++" in s

    r = new_resources(default_code_syntaxes=False)
    syntaxes = r.syntaxes()
    assert not syntaxes


def test_resources_add_syntax():
    r = new_resources(default_code_syntaxes=False)
    r.load_code_syntax_dir(ASSETS_DIR)
    syntaxes = dict(r.syntaxes())
    assert "testC" in syntaxes
    deck = SlideDeck(resources=r)
    slide = deck.new_slide()
    slide.code("if (x > 0) { return 1 }", "testC")


def test_resources_themes(resources):
    themes = resources.themes()
    assert "InspiredGitHub" in themes


def test_resources_add_theme():
    r = new_resources()
    r.load_code_theme_dir(ASSETS_DIR)
    themes = r.themes()
    assert "test" in themes
    deck = SlideDeck(resources=r)
    slide = deck.new_slide()
    slide.code("if (x > 0) { return 1 }", "C", theme="test")



import pytest

from nelsie import InSteps, TextStyle


def test_set_style_error(deck):
    with pytest.raises(Exception, match="^Invalid color: 'xxx'$"):
        deck.set_style("abc", TextStyle(color="xxx"))

    with pytest.raises(Exception, match="^Font 'xxx' not found.$"):
        deck.set_style("abc", TextStyle(font_family="xxx"))

    with pytest.raises(Exception, match="^Invalid color: 'xxx'$"):
        deck.set_style("abc", InSteps({1: TextStyle(), 4: TextStyle(color="xxx")}))


def test_invalid_box(deck):
    with pytest.raises(Exception, match="^Invalid value: xxx$"):
        deck.new_slide().box(width=InSteps({1: 10, 2: "xxx"}))

    with pytest.raises(Exception, match="^Invalid color: 'xxx'$"):
        deck.new_slide().box(bg_color="xxx")


# TODO
# def test_invalid_text_style(deck):
#     deck.new_slide().text("Text", style=TextStyle(color="xxx"))
